/* This file is part of Wyslij-po
   Copyright (C) 2007, 2009 Sergey Poznyakoff

   Wyslij-po is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   Wyslij-po is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with Wyslij-po; if not, write to the Free Software Foundation,
   Inc. 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA */

#include "wyslij-po.h"

static char rbuf[1024];

static size_t io_timeout = 3;
static size_t io_attempts = 3;

static regex_t *pot_regex;

static int
http_stream_wait (mu_stream_t stream, int flags, size_t *attempt)
{
  int rc;
  int oflags = flags;
  struct timeval tv;

  while (*attempt < io_attempts)
    {
      tv.tv_sec = io_timeout;
      tv.tv_usec = 0;
      rc = mu_stream_wait (stream, &oflags, &tv);
      switch (rc) {
      case 0:
	if (flags & oflags)
	  return 0;
	/* FALLTHROUGH */
      case EAGAIN:
      case EINPROGRESS:
	++*attempt;
	continue;
	
      default:
	return rc;
      }
    }
  return ETIMEDOUT;
}

static char *
obstack_s_expand (struct obstack *pstk, const char *tmpl, const char *val)
{
  while (*tmpl)
    {
      if (tmpl[0] == '%' && tmpl[1] == 's')
	{
	  obstack_grow (pstk, val, strlen (val));
	  tmpl += 2;
	}
      else
	{
	  obstack_1grow (pstk, *tmpl);
	  tmpl++;
	}
    }
  obstack_1grow (pstk, 0);
  return obstack_finish (pstk);
}

static int
parse_project_url (mu_url_t url, struct obstack *pstk,
		    char **phost, char **ppath, long *pport)
{
  int status;
  const char *s;
  size_t size;
  char *host;

  status = mu_url_parse (url);
  if (status)
    {
      mu_error ("Error parsing url `%s': %s",
		mu_url_to_string (url), mu_strerror (status));
      return 1;
    }
  
  status = mu_url_sget_scheme (url, &s);
  if (status)
    {
      mu_error ("cannot get protocol from the url `%s': %s",
		mu_url_to_string (url), mu_strerror (status));
      return 1;
    }

  if (strcmp (s, "http"))
    {
      mu_error ("%s: unsupported protocol", s);
      return 1;
    }

  status = mu_url_get_host (url, NULL, 0, &size);
  if (status)
    {
      mu_error ("cannot extract host information from the url `%s': %s",
		mu_url_to_string (url), mu_strerror (status));
      return 1;
    }

  host = obstack_alloc (pstk, size + 1);
  mu_url_get_host (url, host, size + 1, NULL);

  *phost = host;
  s = mu_url_to_string (url);
  obstack_grow (pstk, s, strlen (s) + 1);
  *ppath = obstack_finish (pstk);
  
  mu_url_get_port (url, pport);
    
  return 0;
}

static int
create_project_url (const char *project,  struct obstack *pstk,
		    char **phost, char **ppath, long *pport)
{
  char *url_str = NULL;
  mu_vartab_t vtab;
  mu_url_t url;
  int rc;
  
  mu_vartab_create (&vtab);
  mu_vartab_define (vtab, "domain", project, 0);
  rc = mu_vartab_expand (vtab, tp_url, &url_str);
  mu_vartab_destroy (&vtab);
  if (rc)
    {
      mu_error ("cannot expand string `%s': %s", tp_url, mu_strerror (rc));
      return rc;
    }

  rc = mu_url_create (&url, url_str);
  if (rc)
    {
      mu_error ("Cannot create url from `%s': %s",
		url_str, mu_strerror (rc));
      free (url_str);
      return 1;
    }
  rc = parse_project_url (url, pstk, phost, ppath, pport);
  mu_url_destroy (&url);
  free (url_str);
  return rc;
}
 
static int
read_project_page (const char *project, struct obstack *pstk)
{
  int ret, off = 0;
  mu_stream_t stream;
  size_t nb, size;
  size_t attempt;
  char *host;
  char *path;
  char *cmd;
  long port = 0; /* FIXME: mustn't be long; this should be fixed in MU */
		
  if (create_project_url (project, pstk, &host, &path, &port))
    return 1;

  if (port == 0)
    port = 80;
  
  cmd = obstack_s_expand (pstk, "GET %s HTTP/1.0\r\n\r\n", path);

  if (verbose > 2)
    printf ("Connecting to %s:%ld\n", host, port);
  
  ret = mu_tcp_stream_create (&stream, host, port, MU_STREAM_NONBLOCK);
  if (ret != 0)
    {
      mu_error ("mu_tcp_stream_create: %s", mu_strerror (ret));
      return 1;
    }

  for (attempt = 0; (ret = mu_stream_open (stream)); )
    {
      if ((ret == EAGAIN || ret == EINPROGRESS) && attempt < io_attempts)
	{
	  ret = http_stream_wait(stream, MU_STREAM_READY_WR, &attempt);
	  if (ret == 0)
	    continue;
	}
      mu_error ("mu_stream_open: %s", mu_strerror (ret));
      return 1;
    }

  for (attempt = 0, size = strlen (cmd); size > 0; )
    {
      ret = mu_stream_write (stream, cmd + off, size, 0, &nb);
      if (ret == 0)
	{
	  if (nb == 0)
	    {
	      mu_error("mu_stream_write: wrote 0 bytes");
	      return 1;
	    }
	  off += nb;
	  size -= nb;
	}
      else if (ret == EAGAIN)
        {
	  if (attempt < io_attempts)
	    {
	      ret = http_stream_wait (stream, MU_STREAM_READY_WR, &attempt);
	      if (ret)
		{
		  mu_error ("http_wait failed: %s", mu_strerror (ret));
		  return 1;
		}
	      continue;
	    }
	  else
	    {
	      mu_error ("mu_stream_write timed out");
	      return 1;
	    }
	}
      else
	{
	  mu_error ("mu_stream_write: %s", mu_strerror (ret));
	  return 1;
	}
    }

  if (verbose > 2)
    printf ("Getting %s\n", path);

  attempt = 0;
  for (;;)
    {
      ret = mu_stream_read (stream, rbuf, sizeof (rbuf), 0, &nb);
      if (ret == 0)
	{
	  if (nb == 0)
	    break;
	  obstack_grow (pstk, rbuf, nb);
	}
      else if (ret == EAGAIN)
	{
	  if (attempt < io_attempts)
	    {
	      ret = http_stream_wait (stream, MU_STREAM_READY_RD, &attempt);
	      if (ret)
		{
		  mu_error ("http_stream_wait failed: %s", mu_strerror (ret));
		  return 1;
		}
	    }
	  else
	    {
	      mu_error ("mu_stream_read: %s", mu_strerror (ret));
	      return 1;
	    }
	}
    }

  ret = mu_stream_close (stream);
  if (ret != 0)
    mu_error ("mu_stream_close: %s", mu_strerror (ret));

  mu_stream_destroy (&stream, NULL);
  return 0;
}

static int
http_ok (const char *text)
{
  int code;
  
  if (verbose > 2)
    printf ("Got %-12.12s\n", text);
      
  /* A minimal response is 12 bytes long:
        HTTP/1.1 200
  */	
  if (strlen (text) < 12)
    {
      mu_error ("Invalid HTTP response (got %s)", text);
      return 1;
    }
  
  if (memcmp (text, "HTTP/", 5)
      || !isdigit (text[5]) || text[6] != '.' || !isdigit (text[7])
      || !ISHSPACE (text[8]))
    {
      mu_error ("Invalid HTTP response (got %*.*s)", 8, 8, text);
      return 1;
    }
  for (text += 8; *text; text++)
    if (!ISHSPACE (*text))
      break;

  if (*text == 0)
    {
      mu_error ("Invalid HTTP response: not enough data");
      return 1;
    }

  if (sscanf (text, "%3d", &code) != 1)
    {
      mu_error ("Invalid HTTP response (near %s)", text);
      return 1;
    }

  if (code == 404)
    {
      mu_error ("Page not found");
      return 1;
    }

  if (code != 200)
    {
      mu_error ("Got error %d", code);
      return 1;
    }
  
  return 0;
}
  
int
verify_pot_version (const char *project, const char *vers)
{
  int rc;
  struct obstack stk;

  if (!pot_regex)
    return 1;
  
  if (verbose > 1)
    printf ("Verifying POT file version for `%s'\n", project);
  obstack_init (&stk);
  rc = read_project_page (project, &stk);
  if (rc == 0)
    {
      char *text = obstack_finish (&stk);
      regmatch_t match[2];
      
      if (http_ok (text))
	exit (1);
      
      if (regexec (pot_regex, text, sizeof match / sizeof match[0], match, 0))
	{
	  mu_error ("URL of the latest POT file not found");
	  rc = 1;
	}
      else
	{
	  size_t len = match[1].rm_eo - match[1].rm_so;
	  char *package_str, *pot_str;

	  if (len > 4 && memcmp (text + match[1].rm_eo - 4, ".pot", 4) == 0)
	    len -= 4;
	  package_str = obstack_alloc (&stk, len + 1);
	  memcpy (package_str, text + match[1].rm_so, len);
	  package_str[len] = 0;

	  len = strlen (project) + 1 + strlen (vers);
	  pot_str = obstack_alloc (&stk, len + 1);
	  strcpy (pot_str, project);
	  strcat (pot_str, "-");
	  strcat (pot_str, vers);
	  if (strcmp (package_str, pot_str))
	    {
	      mu_error ("POT project string (%s) does not match latest version (%s)", pot_str, package_str);
	      exit (1);
	    }
	}
    }
  obstack_free (&stk, NULL);
  return rc;
}

void
compile_pot_regex_str ()
{
  regex_t re;
  int rc = regcomp (&re, pot_regex_str, REG_EXTENDED);
  if (rc)
    {
      regerror(rc, &re, rbuf, sizeof rbuf);
      mu_error ("Cannot compile regex: %s", rbuf);
      exit(1);
    }
  if (re.re_nsub < 1)
    {
      mu_error ("Invalid regex: no subexpressions");
      exit(1);
    }
  pot_regex = xmalloc (sizeof (pot_regex[0]));
  *pot_regex = re;
}
